#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# Communicator Architecture
# =========================
#
# This communicator allows us to track the sending of multiple requests and match their responses, which may be streams
# of multiple results, without blocking the initial request call and without any complex threading. This allows the
# usage of a single stream from a single thread to have predictable results, as there will never be race conditions
# between requests.
#
# The result is that requests are:
# * Async: like promises.
# * Serial: always made in the order they are requested by the caller thread.
#
# When a request is made, a Resolver is pushed onto the `resolver_queue`. The Resolver is used to *pull* the correct
# responses from the GRPC stream, which must arrive in a corresponding order but could be pulled by the client in any
# order. There is no "push" from GRPC, the GRPC stream is pulled until the correct responses are found.
#
# No matter which resolver is being resolved, the resolvers *must* be iterated in the order  of the `resolver_queue`,
# since that wil be the order of the responses. If another resolver is ahead of the one we want to retrieve responses
# for, any responses we find are pushed to its buffer, so that when the user later retrieves responses through that
# resolver, it will not miss any. **This push only happens when trying to retrieve results out of the order you request
# them**, it does not happen in a separate thread!
#
# As a final step, closing the transaction must ensure that at least one valid response is received for each resolver,
# to catch and raise any errors before returning control to the user code.


import six
from collections import deque
from six.moves import queue
from grakn.exception.GraknError import GraknError


class SingleResolver:
    def __init__(self, communicator, request):
        self._request = request
        self._communicator = communicator
        self._buffered_response = None
        communicator._send_with_resolver(self, request)

    def _is_last_response(self, response):
        return True

    def _buffer_response(self, response):
        self._buffered_response = response

    def get(self):
        response = self._buffered_response
        if response:
            return response
        else:
            response = self._communicator._block_for_next(self)
            self._buffered_response = response
            return response

    def _end(self):
        pass

    def _on_close(self):
        # Ensure any error is correctly raised by waiting for the response even if we are closing
        self.get()


class IterationResolver(six.Iterator):
    def __init__(self, communicator, request, is_last_response):
        self._request = request
        self._is_last_response = is_last_response
        self._communicator = communicator
        self._ended = False
        self._started = False
        self._response_buffer = deque()
        communicator._send_with_resolver(self, request)

    def __iter__(self):
        return self

    def __next__(self):
        # Regardless of the outcome, we must have a result or error before this method returns
        self._started = True
        try:
            # We should first return any results that another resolver has buffered for us
            return self._response_buffer.popleft()
        except IndexError:
            if self._ended:
                raise StopIteration()
            # Block on the GRPC stream and returns the next result, not requiring the buffer
            # This may buffer results for other resolvers
            return self._communicator._block_for_next(self)

    def _buffer_response(self, response):
        # Only called when another resolver is pushing a result into our buffer, this may happen before first __next__
        self._started = True
        self._response_buffer.append(response)

    def _end(self):
        self._ended = True

    def _on_close(self):
        try:
            while not self._ended:
                response = self._communicator._block_for_next(self)
                self._buffer_response(response)
        except StopIteration:
            pass

    def get(self):
        if self._started:
            return self
        try:
            self._response_buffer.append(next(self))
        except StopIteration:
            pass


class Communicator(six.Iterator):
    def __init__(self, grpc_stream_constructor):
        self._request_queue = queue.Queue()
        self._resolver_queue = deque()
        self._response_iterator = grpc_stream_constructor(self)
        self._closed = False
        self._error = None

    def __iter__(self):
        return self

    # Used by the GRPC stream to iterate requests as they arrive
    def __next__(self):
        request = self._request_queue.get(block=True)
        if request is None:
            raise StopIteration()
        return request

    def error_if_closed(self):
        if self._closed:
            raise GraknError("This connection is closed")

    # Put a request for GRPC to consume
    def _send_with_resolver(self, resolver, request):
        self._resolver_queue.append(resolver)
        self._request_queue.put(request)

    def _block_for_next(self, resolver):
        self.error_if_closed()
        while True:
            current = None
            try:
                current = self._resolver_queue[0]
                response = next(self._response_iterator)

                if current._is_last_response(response):
                    self._resolver_queue.popleft()
                    current._end()

                if current is resolver:
                    return response
                else:
                    current._buffer_response(response)

            except Exception as e:
                if not current:
                    self._error = GraknError("Internal client/protocol error, request/response pair not matched: {0}\n\n "
                                     "Ensure client version is compatible with server version.".format(e))
                else:
                    self._error = GraknError("Server/network error: {0}\n\n generated from request: {1}".format(e, current._request))
                self.close()
                raise self._error

    def iteration_request(self, request, is_last_response):
        self.error_if_closed()
        return IterationResolver(self, request, is_last_response)

    def single_request(self, request):
        self.error_if_closed()
        return SingleResolver(self, request).get()

    def close(self):
        if not self._closed:
            raise_error = False
            if not self._error:
                # Exhaust all resolvers, ensuring that transaction is closed without error
                try:
                    for resolver in list(self._resolver_queue):
                        resolver._on_close()
                except GraknError as e:
                    self._error = e
                    raise_error = True
            self._closed = True

            with self._request_queue.mutex:  # probably don't even need the mutex
                self._request_queue.queue.clear()
            self._request_queue.put(None)

            # force exhaust the iterator so `onCompleted()` is called on the server
            if not self._error:
                try:
                    next(self._response_iterator)
                except StopIteration:
                    pass

            if raise_error:
                raise self._error



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#


from grakn.service.Session.Concept import BaseTypeMapping


def base_type_of(concept):
    return object_to_name[type(concept)]


class Concept(object):

    def __init__(self, grpc_concept):
        self.id = grpc_concept.id
        self.base_type = base_type_of(self)

    def as_remote(self, tx):
        from grakn.service.Session.Concept import ConceptFactory
        return ConceptFactory.create_remote_concept_base(tx._tx_service, self.id, BaseTypeMapping.name_to_grpc_base_type[base_type_of(self)])

    def is_schema_concept(self):
        """ Check if this concept is a schema concept """
        return isinstance(self, SchemaConcept)

    is_schema_concept.__annotations__ = {'return': bool}

    def is_type(self):
        """ Check if this concept is a Type concept """
        return isinstance(self, Type)

    is_type.__annotations__ = {'return': bool}

    def is_thing(self):
        """ Check if this concept is a Thing concept """
        return isinstance(self, Thing)

    is_thing.__annotations__ = {'return': bool}

    def is_attribute_type(self):
        """ Check if this concept is an AttributeType concept """
        return isinstance(self, AttributeType)

    is_attribute_type.__annotations__ = {'return': bool}

    def is_entity_type(self):
        """ Check if this concept is an EntityType concept """
        return isinstance(self, EntityType)

    is_entity_type.__annotations__ = {'return': bool}

    def is_relation_type(self):
        """ Check if this concept is a RelationType concept """
        return isinstance(self, RelationType)

    is_relation_type.__annotations__ = {'return': bool}

    def is_role(self):
        """ Check if this concept is a Role """
        return isinstance(self, Role)

    is_role.__annotations__ = {'return': bool}

    def is_rule(self):
        """ Check if this concept is a Rule concept """
        return isinstance(self, Rule)

    is_rule.__annotations__ = {'return': bool}

    def is_attribute(self):
        """ Check if this concept is an Attribute concept """
        return isinstance(self, Attribute)

    is_attribute.__annotations__ = {'return': bool}

    def is_entity(self):
        """ Check if this concept is an Entity concept """
        return isinstance(self, Entity)

    is_entity.__annotations__ = {'return': bool}

    def is_relation(self):
        """ Check if this concept is a Relation concept """
        return isinstance(self, Relation)

    is_relation.__annotations__ = {'return': bool}


class SchemaConcept(Concept):

    def __init__(self, grpc_concept):
        super(SchemaConcept, self).__init__(grpc_concept)
        self._label = grpc_concept.label_res.label

    def label(self):
        """
        Get the label of this schema concept.
        """
        return self._label


class Type(SchemaConcept):
    pass


class EntityType(Type):
    pass


class AttributeType(Type):

    def __index__(self, grpc_concept):
        super(Type, self).__init__(grpc_concept)
        from grakn.service.Session.util import ResponseReader
        self._value_type = ResponseReader.ResponseReader.from_grpc_value_type_res(grpc_concept.valueType_res)

    def value_type(self):
        """ Get the ValueType enum (grakn.ValueType) corresponding to the type of this attribute """
        return self._value_type


class RelationType(Type):
    pass


class Rule(SchemaConcept):
    pass


class Role(SchemaConcept):
    pass


class Thing(Concept):

    def __init__(self, grpc_concept):
        super(Thing, self).__init__(grpc_concept)
        self._inferred = grpc_concept.inferred_res.inferred
        from grakn.service.Session.Concept import ConceptFactory
        self._type = ConceptFactory.create_local_concept(grpc_concept.type_res.type)

    def is_inferred(self):
        return self._inferred

    def type(self):
        return self._type


class Entity(Thing):
    pass


class Attribute(Thing):

    def __init__(self, grpc_concept):
        super(Attribute, self).__init__(grpc_concept)
        from grakn.service.Session.util import ResponseReader
        self._value = ResponseReader.ResponseReader.from_grpc_value_object(grpc_concept.value_res.value)

    def value(self):
        return self._value


class Relation(Thing):
    pass


name_to_object = {
    BaseTypeMapping.META_TYPE: Type,
    BaseTypeMapping.ENTITY_TYPE: EntityType,
    BaseTypeMapping.RELATION_TYPE: RelationType,
    BaseTypeMapping.ATTRIBUTE_TYPE: AttributeType,
    BaseTypeMapping.ROLE: Role,
    BaseTypeMapping.RULE: Rule,
    BaseTypeMapping.ENTITY: Entity,
    BaseTypeMapping.RELATION: Relation,
    BaseTypeMapping.ATTRIBUTE: Attribute
}

object_to_name = dict(zip(name_to_object.values(), name_to_object.keys()))



#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
# 
#   http://www.apache.org/licenses/LICENSE-2.0
# 
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

import six
from six.moves import map

from grakn.service.Session.util.RequestBuilder import RequestBuilder
import grakn.service.Session.util.ResponseReader as ResponseReader # for circular import issue
from grakn.service.Session.util import enums
from grakn.service.Session.util.Communicator import Communicator
from grakn.exception.GraknError import GraknError


class TransactionService(object):

    def __init__(self, session_id, tx_type, transaction_endpoint):
        self.session_id = session_id
        self.tx_type = tx_type.value

        self._communicator = Communicator(transaction_endpoint)

        # open the transaction with an 'open' message
        open_req = RequestBuilder.open_tx(session_id, tx_type)
        self._communicator.single_request(open_req)
    __init__.__annotations__ = {'tx_type': enums.TxType}

    # --- Passthrough targets ---
    # targets of top level Transaction class

    def query(self, query, infer, explain, batch_size):
        return Iterator(self._communicator,
                        RequestBuilder.start_iterating_query(query, infer, explain, batch_size),
                        ResponseReader.ResponseReader.get_query_results(self))
    query.__annotations__ = {'query': str}

    def commit(self):
        request = RequestBuilder.commit()
        self._communicator.single_request(request)

    def close(self):
        self._communicator.close()

    def is_closed(self):
        return self._communicator._closed

    def get_concept(self, concept_id):
        request = RequestBuilder.get_concept(concept_id)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.get_concept(self, response.getConcept_res)
    get_concept.__annotations__ = {'concept_id': str}

    def get_schema_concept(self, label):
        request = RequestBuilder.get_schema_concept(label)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.get_schema_concept(self, response.getSchemaConcept_res)
    get_schema_concept.__annotations__ = {'label': str}

    def get_attributes_by_value(self, attribute_value, value_type):
        request = RequestBuilder.start_iterating_get_attributes_by_value(attribute_value, value_type)
        return Iterator(self._communicator, request, ResponseReader.ResponseReader.get_attributes_by_value(self))
    get_attributes_by_value.__annotations__ = {'data_type': enums.ValueType}

    def put_entity_type(self, label):
        request = RequestBuilder.put_entity_type(label)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.put_entity_type(self, response.putEntityType_res)
    put_entity_type.__annotations__ = {'label': str}

    def put_relation_type(self, label):
        request = RequestBuilder.put_relation_type(label)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.put_relation_type(self, response.putRelationType_res)
    put_relation_type.__annotations__ = {'label': str}

    def put_attribute_type(self, label, value_type):
        request = RequestBuilder.put_attribute_type(label, value_type)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.put_attribute_type(self, response.putAttributeType_res)
    put_attribute_type.__annotations__ = {'label': str, 'value_type': enums.ValueType}

    def put_role(self, label):
        request = RequestBuilder.put_role(label)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.put_role(self, response.putRole_res)
    put_role.__annotations__ = {'label': str}

    def put_rule(self, label, when, then):
        request = RequestBuilder.put_rule(label, when, then)
        response = self._communicator.single_request(request)
        return ResponseReader.ResponseReader.put_rule(self, response.putRule_res)
    put_rule.__annotations__ = {'label': str, 'when': str, 'then': str}

    # --- Transaction Messages ---

    def run_concept_method(self, concept_id, grpc_concept_method_req):
        # wrap method_req into a transaction message
        tx_request = RequestBuilder.concept_method_req_to_tx_req(concept_id, grpc_concept_method_req)
        response = self._communicator.single_request(tx_request)
        return response.conceptMethod_res.response

    def run_concept_iter_method(self, concept_id, grpc_concept_iter_method_req):
        return Iterator(self._communicator,
                        RequestBuilder.start_iterating_concept_method(concept_id, grpc_concept_iter_method_req),
                        lambda res: res.conceptMethod_iter_res.response)

    def explanation(self, explainable):
        """ Retrieve the explanation of a Concept Map from the server """
        tx_request = RequestBuilder.explanation(explainable)
        response = self._communicator.single_request(tx_request)
        return ResponseReader.ResponseReader.create_explanation(self, response.explanation_res)


end_of_batch_results = {'done', 'iteratorId'}


def end_of_batch(res):
    res_type = res.iter_res.WhichOneof('res')
    return res_type == 'done' or res_type == 'iteratorId'


class Iterator(six.Iterator):
    def __init__(self, communicator, iter_req, resp_converter):
        self._communicator = communicator
        self._iter_req = iter_req
        self._response_iterator = self._communicator.iteration_request(
            RequestBuilder.iter_req_to_tx_req(self._iter_req),
            end_of_batch)
        self._done = False
        self._resp_converter = resp_converter

    def __iter__(self):
        return self

    def _request_next_batch(self, iter_id):
        self._response_iterator = self._communicator.iteration_request(
            RequestBuilder.iter_req_to_tx_req(
                RequestBuilder.continue_iterating(iter_id, self._iter_req.options)),
            end_of_batch)

    def __next__(self):
        if self._done:
            raise GraknError('Iterator was already iterated.')

        try:
            response = next(self._response_iterator)
        except StopIteration:
            raise GraknError('Internal client/protocol error,'
                             ' did not receive an expected "done" or "iteratorId" message.'
                             '\n\n Please ensure client version is supported by server version.')

        iter_res = response.iter_res
        res_type = iter_res.WhichOneof('res')
        if res_type == 'done':
            self._done = True
            raise StopIteration
        elif res_type == 'iteratorId':
            self._request_next_batch(iter_res.iteratorId)
            return next(self)
        else:
            return self._resp_converter(iter_res)

    def get(self):
        if not self._done:
            self._response_iterator.get()
        return self
